from pygame import mixer

from pynput.keyboard import Listener

from keyboardsounds.profile import Profile
from keyboardsounds.audio_manager import AudioManager

__am: AudioManager = None
__volume = 100
__down = []

def __on_press(key):
    global __am
    global __volume
    global __down

    if key in __down:
        return

    sound = __am.get_sound(key, action="press")
    if sound is not None:
        clip = mixer.Sound(sound)
        clip.set_volume(float(__volume) / float(100))
        clip.play()

    # Play the sound
    __down.append(key)

def __on_release(key):
    global __down
    global __am

    sound = __am.get_sound(key, action="release")
    if sound is not None:
        clip = mixer.Sound(sound)
        clip.set_volume(float(__volume) / float(100))
        clip.play()

    __down = [k for k in __down if k != key]

def run(volume: int, profile: str):
    global __am
    global __volume

    __volume = volume
    __am = AudioManager(Profile(profile))

    mixer.init()
    with Listener(on_press=__on_press, on_release=__on_release) as listener:
        listener.join()
        


import os
import sys
import psutil
import subprocess
import time
import json

import keyboardsounds.daemon as daemon
from keyboardsounds.profile import Profile

class DaemonManager:
    def __init__(self, lock_file) -> None:
        self.__lock_file = lock_file
        self.__proc_info = None
        self.__lock_exists = False
        self.__is_daemon_process = False
        self.__proc = None
        self.__load_status()

    def __load_status(self):
        if os.path.isfile(self.__lock_file):
            self.__lock_exists = True
            try:
                with open(self.__lock_file, "r") as f:
                    self.__proc_info = json.load(f)
            except ValueError:
                self.__proc_info = None
                pass
        else:
            self.__proc_info = None

        if self.__proc_info:
            try:
                self.__proc = psutil.Process(self.__proc_info["pid"])
            except psutil.NoSuchProcess:
                self.__proc = None
                pass
        else:
            self.__proc = None

        if self.__proc:
            try:
                self.__is_daemon_process = self.__proc.name() == psutil.Process().name()
            except ValueError:
                pass
        else:
            self.__is_daemon_process = False

    def status(self, full=False) -> str:
        if full:
            volume = self.get_volume()
            volume_status = f", Volume: {volume}%" if volume is not None else ""
            pid = self.get_pid()
            pid_status = f", PID: {pid}" if pid is not None else ""
            profile = self.get_profile()
            profile_status = f", Profile: {profile}" if profile is not None else ""
            status_text = "Running" if self.status() == "running" else "Stale" if self.status() == "stale" else "Not running"
            return f"Status: {status_text}{volume_status}{pid_status}{profile_status}"
        else:
            self.__load_status()
            if self.__lock_exists:
                if self.__is_daemon_process:
                    return "running"
                else:
                    return "stale"
            else:
                return "free"

    def get_volume(self) -> int:
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["volume"]
        return None

    def get_pid(self) -> int:
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["pid"]
        return None

    def get_profile(self) -> str:
        self.__load_status()
        status = self.status()
        if status == "running":
            return self.__proc_info["profile"]
        return None

    def try_stop(self) -> None:
        status = self.status()
        if status == "free":
            return False

        if status == "running":
            self.__proc.kill()
            status = self.status()

        if status == "stale":
            os.unlink(self.__lock_file)

        self.__proc_info = None
        self.__lock_exists = False
        self.__is_daemon_process = False
        self.__proc = None

        return True

    def try_start(self, volume: int, profile: str) -> None:
        try:
            Profile(profile)
        except ValueError as err:
            print(F"Error: {err}")
            return False

        status = self.status()
        if status == "running":
            self.try_stop()
            status = self.status()

        if status == "stale":
            self.try_stop()

        subprocess.Popen(
            [sys.argv[0], "start-daemon", str(volume), profile],
            creationflags=subprocess.CREATE_NO_WINDOW if os.name == "nt" else 0,
            start_new_session=True
        )
        time.sleep(1.0)
        return True

    def capture_daemon_initialization(self):
        if len(sys.argv) == 4 and sys.argv[1] == "start-daemon":
            if self.status() == "running":
                return

            volume = 100
            try:
                volume = int(sys.argv[2])
            except:
                pass

            profile = "ios"
            try:
                profile = sys.argv[3]
            except:
                pass

            with open(self.__lock_file, 'w') as f:
                json.dump({
                    "pid": os.getpid(),
                    "volume": volume,
                    "profile": profile,
                }, f)

            f = open(os.devnull, 'w')
            sys.stdout = f
            sys.stderr = f
            daemon.run(volume, profile)
            return True
        return False
        


import subprocess
import wave
import os
import io
import random

from imageio_ffmpeg import get_ffmpeg_exe
from pynput.keyboard import Key, KeyCode

from keyboardsounds.profile import Profile

class AudioManager:
    def __init__(self, profile: Profile) -> None:
        self.sounds = {}
        self.profile = profile
        self.__prime_audio_clips()

    def __prime_audio_clips(self):
        if self.profile.value('profile.type') == "video-extract":
            ffmpeg_exe = get_ffmpeg_exe()
            VIDEO_FILE = self.profile.get_file_path(self.profile.value('profile.video'))
            AUDIO_FILE = f"{VIDEO_FILE}.wav"

            # Convert video file to wav
            subprocess.run(
                [
                    ffmpeg_exe,
                    "-y",
                    "-i",
                    VIDEO_FILE,
                    "-f",
                    "wav",
                    "-vn",
                    AUDIO_FILE
                ],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )

            for source in self.profile.value('sources'):
                self.__extract(source["id"], AUDIO_FILE, source["start"], source["end"])

            os.unlink(AUDIO_FILE)

        elif self.profile.value('profile.type') == "files":
            for source in self.profile.value('sources'):
                if type(source["source"]) == dict:
                    self.__extract(f"{source['id']}__press", self.profile.get_file_path(source["source"]["press"]))
                    self.__extract(f"{source['id']}__release", self.profile.get_file_path(source["source"]["release"]))
                    self.sounds[source['id']] = {
                        "press": io.BytesIO(self.sounds[f"{source['id']}__press"].getbuffer().tobytes()),
                        "release": io.BytesIO(self.sounds[f"{source['id']}__release"].getbuffer().tobytes())
                    }
                    del self.sounds[f"{source['id']}__press"]
                    del self.sounds[f"{source['id']}__release"]
                elif type(source["source"]) == str:
                    self.__extract(source["id"], self.profile.get_file_path(source["source"]))

    def __extract(self, id, input, start: float = 0.0, end: float = None):
        if input.endswith(('mp3', 'MP3')):
            source = open(input, 'rb')
            data = source.read()
            source.close()
            self.sounds[id] = io.BytesIO(data)
        else:
            source = wave.open(input, 'rb')
            source.setpos(int(start * source.getframerate()))
            end = end if end is not None else source.getnframes() / source.getframerate()
            frames = int((end - start) * source.getframerate())
            self.sounds[id] = io.BytesIO()
            dest = wave.open(self.sounds[id], 'wb')
            dest.setparams(source.getparams())
            dest.writeframes(source.readframes(frames))
            source.close()
            self.sounds[id].seek(0)

    def get_sound(self, key, action: str = "press") -> io.BytesIO:
        if self.profile.value('keys') is not None:
            if self.profile.value('keys.other') is not None:
                for mapping in self.profile.value('keys.other'):
                    if (key.name if isinstance(key, Key) else key.char) in mapping["keys"]:
                        return self.__get_sound(mapping["sound"], action)
            if self.profile.value('keys.default') is not None:
                return self.__get_sound(self.profile.value('keys.default'), action)
        return self.__get_sound(key=None, action=action)

    def __get_sound(self, key=None, action: str = "press") -> io.BytesIO:
        if key is None:
            key = list(self.sounds.keys())
        if type(key) is list:
            return self.__parse_sound(self.sounds[random.choice(key)], action)
        return self.__parse_sound(self.sounds[key], action)
    
    def __parse_sound(self, sound, action: str = "press") -> io.BytesIO:
        if type(sound) is dict:
            return io.BytesIO(sound[action].getbuffer().tobytes())
        return io.BytesIO(sound.getbuffer().tobytes()) if action == "press" else None
